from senatus.api.binanceAPI import (getPastCandlesTimeSeries)
from senatus.indicators.indicators import (getMAV, getEMA, getWRSI, find_top)
from senatus.plot import Plot
from senatus.test_trader import TestTrader
from senatus.strategies import RSI_Margins_Strategy, BoolIndicatorStrategy

# Define your parameters
pair = 'bnbusdt'
timeframe = '1h'


# Get past data
priceSeries = getPastCandlesTimeSeries(pair, timeframe)


# Plot data
my_plot = Plot(priceSeries)

my_plot.plot_asset_chart()

wrsi = getWRSI(priceSeries, 8)

my_plot.plot_indicator_dif_chart(wrsi)

is_top = find_top(priceSeries, 25, 0)

tt = TestTrader(priceSeries, 1000)

#tt.operate_strategy(is_top, BoolIndicatorStrategy(), my_plot)

#tt.operate_strategy(wrsi, RSI_Margins_Strategy(15, 80), my_plot)

my_plot.apply_custom().show()


import pandas as pd
from ..data import PriceSeries, MetricSeries
from .indicators_utils import (
    calcRSI, 
    calcWRSI,
    eval_top,
)

def getMAV(ps: PriceSeries, period: int = 20):
    df = ps.candles
    return MetricSeries(pd.DataFrame({
        'timestamp': df.index,
        'value': df['close'].rolling(window=period).mean()
    }).set_index('timestamp'), ps.timeframe)

def getEMA(ps: PriceSeries, period: int = 20):
    df = ps.candles
    return MetricSeries(pd.DataFrame({
        'timestamp': df.index,
        'value': df['close'].ewm(span=period, adjust=False).mean()
    }).set_index('timestamp'), ps.timeframe)

def getRSI(ps: PriceSeries, period: int = 14):
    df = ps.candles
    # Ensure 'close' column is numeric
    df['close'] = pd.to_numeric(df['close'], errors='coerce')
    return MetricSeries(pd.DataFrame({
        'timestamp': df.index,
        'value': calcRSI(df['close'], period),
    }), ps.timeframe)

def getWRSI(ps: PriceSeries, period: int = 14):
    df = ps.candles
    # Ensure 'close' column is numeric
    df['close'] = pd.to_numeric(df['close'], errors='coerce')
    return MetricSeries(pd.DataFrame({
        'timestamp': df.index,
        'value': calcWRSI(df['close'], period).values,
    }), ps.timeframe)

def find_top(ps: PriceSeries, window: int = 5, middle_shift: int = 0):
    '''
    Useful to get Short signals at tops
    '''
    df = ps.candles
    # Ensure 'close' column is numeric
    df['close'] = pd.to_numeric(df['close'], errors='coerce')
    return MetricSeries(pd.DataFrame({
        'timestamp': df.index,
        'value': eval_top(ps.candles['close'], window, middle_shift),
    }), ps.timeframe)


import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from ..data import (PriceSeries, MetricSeries, Position)
from ..custom import stdLayout


class Plot():
    """
    A single plot, meant to represent an asset with info, data and analysis methods.
    Every instance methods return the instance itself (exept .show() which is supposed to run last), 
    allowing chain calling
    """

    def __init__(self,
            priceSeries: PriceSeries,
            main_figure_rows: int = 3,
            main_figure_cols: int = 1
        ):

        self.priceSeries = priceSeries
        self.main_figure = make_subplots(
            rows=main_figure_rows, 
            cols=main_figure_cols, 
            shared_xaxes=True, 
            vertical_spacing=0,
            specs=[[{"rowspan": 2}], [None], [{}]]
        )

    def plot_asset_chart(self):
        self.main_figure.add_trace( 
            go.Candlestick(
                x=self.priceSeries.candles.index,
                open=self.priceSeries.candles['open'],
                high=self.priceSeries.candles['high'],
                low=self.priceSeries.candles['low'],
                close=self.priceSeries.candles['close']
            ),
        )
        return self

    def apply_custom(self, layout = stdLayout):
        self.main_figure.update(layout_xaxis_rangeslider_visible=False)
        self.main_figure.update_layout(layout)
        return self

    def show(self):
        self.main_figure.show()

    def plot_indicator_same_chart(self, ind: MetricSeries):
        self.main_figure.add_trace(
            go.Scatter(
                x=ind.series.index,
                y=ind.series['value']
            )  
        )
        return self
    
    def plot_indicator_dif_chart(self, ind: MetricSeries, row=3, col=1):
        self.main_figure.add_trace(
            go.Scatter(
                x=ind.series.index, y=ind.series['value']
            ),
            row=row,
            col=col
        )
        return self

    def plotPosition(self, pos: Position):
        if(pos.is_open):
            self.main_figure.add_trace(
                go.Scatter(
                    x=[pos.open_timeframe], 
                    y=[pos.entry_price], 
                    mode='markers',
                    marker=dict(size=[10], color=('green' if pos.position_side == 'LONG' else 'red'))),
                row=1,
                col=1
            )
        return self

    


from .candlestick import Candlestick
from .price_series import PriceSeries
from .metric_series import MetricSeries
from .position import Position

__all__ = [
    "Candlestick",
    "PriceSeries",
    "MetricSeries",
    "Position",
]


import pandas as pd

class Candlestick:
    """
    A candlestick representation
    """

    def __init__(
        self,
        timestamp: pd.Timestamp,    # Timestamps for date
        open: float,                # Open prices
        high: float,                # High prices
        low: float,                 # Low prices
        close: float,               # Close prices
        volume: float,              # Volume data
    ):
        self.timestamp = timestamp
        self.open = open
        self.high = high
        self.low = low
        self.close = close
        self.volume = volume

    def __repr__(self):
        return f"{self.timestamp} {self.open} {self.high} {self.low} {self.close} {self.volume})"
    
    @classmethod
    def to_dataframe(cls, candlesticks):
        return pd.DataFrame({
            'timestamp': [candle.timestamp for candle in candlesticks],
            'open': [candle.open for candle in candlesticks],
            'high': [candle.high for candle in candlesticks],
            'low': [candle.low for candle in candlesticks],
            'close': [candle.close for candle in candlesticks],
            'volume': [candle.volume for candle in candlesticks] 
        })



import pandas as pd

class Position():
    def __init__(
            self,
            symbol: str,
            is_open: bool,
            type: str,
            position_side: str,
            position_size: float,
            entry_price: float,
            open_timeframe: pd.Timestamp,
            close_timeframe: pd.Timestamp = None,
            close_price: float = None
        ):
        self.symbol = symbol
        self.is_open = is_open
        self.open_timeframe = open_timeframe
        self.close_timeframe = close_timeframe
        self.type = type
        self.position_side = position_side
        self.position_size = position_size
        self.entry_price = entry_price
        self.close_price = close_price

    def get_unrealized_pnl(self):
        unrealized_pnl = (self.current_price - self.entry_price) * self.position_size
        if self.position_side == "SHORT":
            unrealized_pnl *= -1
        return unrealized_pnl
    
    def close_position(self, close_price, close_timeframe):
        self.close_price = close_price
        self.close_timeframe = close_timeframe
        self.is_open = False
        return self
    
    def get_pnl(self):
        if self.is_open: return 0
        pnl = (self.current_price - self.entry_price) * self.position_size
        if self.position_side == "SHORT":
            pnl *= -1
        return pnl





import pandas as pd

class MetricSeries:

    def __init__(
        self,
        data_frame: pd.DataFrame,
        timeframe: str,
    ):  
        self.series = data_frame
        self.timeframe = timeframe
        self.series['timestamp'] = pd.to_datetime(self.series['timestamp'])
        self.series.set_index('timestamp', inplace=True)

    


import pandas as pd
from .candlestick import Candlestick

class PriceSeries:

    def __init__(
        self,
        candlesticks: list[Candlestick],
        pair: str,
        timeframe: str,
    ):  
        # Need to throw an error with invalid candlesticks (like null list)
        self.candles = Candlestick.to_dataframe(candlesticks)
        self.pair = pair
        self.timeframe = timeframe
        self.candles['timestamp'] = pd.to_datetime(self.candles['timestamp'])
        self.candles.set_index('timestamp', inplace=True)

    


import pandas as pd
from .data import Position, MetricSeries
from typing import Callable, Optional


class Strategy:


    conditionToOpenLong: Optional[Callable[[MetricSeries, list[Position]], bool]]
    conditionToCloseLong: Optional[Callable[[MetricSeries, list[Position]], bool]]
    conditionToOpenShort: Optional[Callable[[MetricSeries, list[Position]], bool]]
    conditionToCloseShort:Optional[Callable[[MetricSeries, list[Position]], bool]]
    
    def __init__(
            self, 
            cOL: Optional[Callable[[MetricSeries, list[Position]], bool]] = lambda a, b: False, 
            cCL: Optional[Callable[[MetricSeries, list[Position]], bool]] = lambda a, b: False, 
            cOS: Optional[Callable[[MetricSeries, list[Position]], bool]] = lambda a, b: False, 
            cCS: Optional[Callable[[MetricSeries, list[Position]], bool]] = lambda a, b: False
        ):
        self.conditionToOpenLong = cOL
        self.conditionToCloseLong = cCL
        self.conditionToOpenShort = cOS
        self.conditionToCloseShort = cCS

    def check_condition(self, ind: pd.Series, positions: list[Position] = []):
        resp = []
        if self.conditionToOpenLong(ind, positions): resp.append('OPEN_LONG')
        if self.conditionToCloseLong(ind, positions): resp.append('CLOSE_LONG')
        if self.conditionToOpenShort(ind, positions): resp.append('OPEN_SHORT')
        if self.conditionToCloseShort(ind, positions): resp.append('CLOSE_SHORT')
        return resp
    

class RSI_Margins_Strategy(Strategy):


    def __init__(self, lowerMargin: float = 30, upperMargin: float = 70):
        
        super().__init__(
            cOL = lambda wrsi, na: (wrsi['value'].iat[-1] <= lowerMargin) if not wrsi.empty else False, 
            cCL = lambda wrsi, na: (wrsi['value'].iat[-1] >= upperMargin) if not wrsi.empty else False,
        )

class BoolIndicatorStrategy(Strategy):

    def __init__(self):
        super().__init__(
            cOS = lambda is_top, na: (is_top['value'].iat[-1]) if not is_top.empty else False
        )


from plotly.graph_objects import (Figure, Layout)
from plotly.graph_objs.layout import (XAxis, Margin)

stdLayout = Layout(
    title='btcudst',
    paper_bgcolor='#0a0a0a',
    plot_bgcolor='#0a0a0a',
    margin={
        'b': 16, 'l': 16, 'pad': 0, 'r': 16, 't': 32
    },
    xaxis = {
        'showline' : True,
        'linecolor' : '#282828',
        'linewidth' : 4,
        'gridcolor' : '#282828',
        'gridwidth' : 1,
        'mirror': True,
    },
    yaxis = {
        'showline' : True,
        'linecolor' : '#282828',
        'linewidth' : 4,
        'gridcolor' : '#282828',
        'gridwidth' : 1,
        'mirror': True,
        'side': 'right',
    },
    xaxis2={
        'showline': True,
        'linecolor': '#282828',
        'linewidth': 4,
        'gridcolor': '#282828',
        'gridwidth': 1,
        'mirror': True,
        'anchor': 'y2'  # Tells this axis to align with y2
    },
    yaxis2={
        'showline': True,
        'linecolor': '#282828',
        'linewidth': 4,
        'gridcolor': '#282828',
        'gridwidth': 1,
        'mirror': True,
        'side': 'right',
        'anchor': 'x2'  # Tells this axis to align with x2
    }
    
)


import numpy as np
import pandas as pd

def calcSMA(data: pd.Series, period: int):
    return data.rolling(window=period).mean() 

def calcRSI(close: pd.Series, window: int = 14): 
        delta = close.diff()
        avg_gain = calcSMA(delta.where(delta >= 0, 0))
        avg_loss = calcSMA(-delta.where(delta < 0, 0))
        rs = avg_gain / avg_loss
        rsi = 100 - (100 / (1 + rs))
        return rsi

def calcSMMA(data: pd.Series, period: int):
    #Evaluate the SMMA for each data point
    def evalSMMA(index, value, prev, period):
        if index < period - 1:
            return 0
        if index == period - 1:
            return data[:period].mean()
        return (prev * (period - 1) + value) / period
    smma = pd.Series(dtype=float)
    for i in range(len(data)):
         prev = smma[i - 1] if i > 0 else 0
         smma.loc[i] = (evalSMMA(i, data.iloc[i], prev, period))
    return smma

def calcWRSI(data: pd.Series, period: int):
    #Calculates the Wilder's RSI with his smoothing method (RSI used in Tranding View)
    delta = data.diff()
    ups = delta.where(delta >= 0, 0)
    downs = abs(delta.where(delta < 0, 0))     #Loss values should be treated as positive for calculation
    smmaU = calcSMMA(ups, period)
    smmaD = calcSMMA(downs, period)
    rs = smmaU.div(smmaD.replace(0, np.nan), fill_value=0)  # Avoid division by zero
    return rs.apply(lambda elem: 100 - (100/(1+elem)))

def eval_top(data: pd.Series, window: int, middle_shift: int):
    top_indicator = pd.Series(dtype=bool)
    data = data.reset_index(drop=True)
    for index, item in data.items():
        if(index < window - middle_shift):
            top_indicator.loc[index] = False
            pass
        lower = index - int(np.ceil(window/2)) - middle_shift
        upper = index + int(np.floor(window/2)) + middle_shift
        shifted_window = data.iloc[lower:upper].replace(item, np.nan)
        max_neighbor = shifted_window.max()
        top_indicator.loc[index] = (item > max_neighbor)
    return top_indicator


import pandas as pd
from binance.um_futures import UMFutures
from ..data import Candlestick, PriceSeries

def processRawCandlestickData(rawCD: list) -> list[Candlestick]:
    """
    Transforms raw data from Binance kline API to Candlestick type
    """
    
    return [
        Candlestick(
            timestamp = pd.Timestamp(cd[0], unit='ms'),
            open = cd[1],
            high = cd[2],
            low = cd[3],
            close = cd[4],
            volume = cd[5],
        ) for cd in rawCD
    ]


def getPastCandlesTimeSeries(pair: str, timeframe: str) -> PriceSeries:
    """
    Request Binance API kline data and returns a Pandas DataFrame with Candlesticks
    """

    um_futures_client = UMFutures()
    pastCandleSticks = um_futures_client.klines(pair, timeframe, limit=1000)
    priceSeries = PriceSeries(processRawCandlestickData(pastCandleSticks), pair, timeframe)
    return priceSeries


from .data import PriceSeries, MetricSeries, Position
from .plot import Plot
from .strategies import Strategy

class TestTrader:

    def __init__( 
            self,
            ps: PriceSeries,
            initial_margin: float
        ):
        self.df = ps.candles
        self.initial_margin = initial_margin
        self.open_positions: list[Position] = []

    def open_position(self, new_position: Position):
        self.open_positions.append(new_position)
        return self.open_positions[-1]

    def close_position(self, p_timeframe):
        return next((pos for pos in self.open_positions if pos.open_timeframe == p_timeframe), None)
    
    def operate_strategy(self, ind: MetricSeries, strategy: Strategy, plot: Plot):
        for i in range(len(self.df)):
            resp = strategy.check_condition(ind.series[0:i])
            if(resp.count('OPEN_LONG')):
                pos = self.open_position(Position(
                    symbol=plot.priceSeries.pair,
                    is_open=True,
                    type='spot',
                    position_side='LONG',
                    position_size=1,
                    entry_price=self.df['close'].iat[i],
                    open_timeframe=self.df.index[i]
                ))
                plot.plotPosition(pos)
            if(resp.count('CLOSE_LONG')):
                pos = self.open_position(Position(
                    symbol=plot.priceSeries.pair,
                    is_open=True,
                    type='spot',
                    position_side='SHORT',
                    position_size=1,
                    entry_price=self.df['close'].iat[i],
                    open_timeframe=self.df.index[i]
                ))
                plot.plotPosition(pos)
            if(resp.count('OPEN_SHORT')):
                pos = self.open_position(Position(
                    symbol=plot.priceSeries.pair,
                    is_open=True,
                    type='spot',
                    position_side='SHORT',
                    position_size=1,
                    entry_price=self.df['close'].iat[i],
                    open_timeframe=self.df.index[i]
                ))
                plot.plotPosition(pos)
            if(resp.count('CLOSE_SHORT')):
                pos = self.open_position(Position(
                    symbol=plot.priceSeries.pair,
                    is_open=True,
                    type='spot',
                    position_side='LONG',
                    position_size=1,
                    entry_price=self.df['close'].iat[i],
                    open_timeframe=self.df.index[i]
                ))
                plot.plotPosition(pos)
        
